/*
 * OptiController.java
 *
 * Created on 4. desember 2006, 11:59
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package CoreSys;

/**
 *
 * @author Sindre Soltun
 */
public class OptiController {
    
    /** Creates a new instance of OptiController */
    public OptiController() {
    }
    
    /** Runs job otimisation */
    public void optimiseJobs(int[][] costs, int travellers, int demand[],int capacity){
        int[] farthest = farthestInsertion(costs,travellers);
        
        int[] clarkewright = clarkeWright(costs,travellers,demand,capacity);
        int lengthF = 0;
        int lengthCF = 0;
        for(int i = 0 ; i < farthest.length - 1 ; i++){
            lengthF += costs[farthest[i]][farthest[i+1]];
            lengthCF += costs[clarkewright[i]][clarkewright[i+1]];
        }
        System.out.println("Total length FI: "+lengthF);
        System.out.println("Total length CW: "+lengthCF);
        System.out.println("");
    }
    /** Creates savings matrix for Clarke & Wright algorithm */
    public int[][] createSavingsMatrix(int[][] matrix){
        int[][] savingsmatrix = new int[matrix.length - 1][matrix.length - 1];
        for(int i = 0 ; i < savingsmatrix.length ; i++){
            savingsmatrix[i][i] = 0;
            for(int j = i + 1 ; j < savingsmatrix.length ; j++){
                savingsmatrix[i][j] = matrix[0][i+1] + matrix[0][j+1] - matrix[i+1][j+1];
                savingsmatrix[j][i] = savingsmatrix[i][j];
            }
        }

        return savingsmatrix;
    }
    /** Runs Clarke & Wright algorithm based on given input */
    public int[] clarkeWright(int[][] m, int travellers, int[] demand, int cap){
        int[][] sm = createSavingsMatrix(m);
        int nodeA = 0;
        int nodeB = 0;
        int endA = 0;
        int endB = 0;
        int[][] tour = new int[sm.length][m.length + 1];
        boolean terminate = false;
        boolean[] finished = new boolean[sm.length];
        boolean[] active = new boolean[sm.length];
        int[] belongs = new int[sm.length];
        int[] tourcost = new int[sm.length];
        for(int i = 0 ; i < sm.length ; i++){
            for(int j = 0 ; j <= m.length ; j++){
                if(j == 1){
                    tour[i][j] = i+1;
                }
                else{
                    tour[i][j] = 0;
                }
            }
        }
        for(int i = 0 ; i < sm.length ; i++){
            finished[i] = false;
            active[i] = false;
            belongs[i] = i;
            tourcost[i] = demand[i];
        }
        boolean intour = false;
        
        int carsused = 0;
        while(!terminate){
            int bestcurrentsaving = -10000;
            if(!intour){
                for(int i = 0 ; i < sm.length ; i++){
                    if(!finished[belongs[i]]){
                        for(int j = i + 1 ; j < sm.length ; j++){
                            if(!finished[belongs[j]] && sm[i][j] > bestcurrentsaving && tourcost[belongs[i]] + demand[j] <= cap){
                                nodeA = i;
                                nodeB = j;
                                bestcurrentsaving = sm[i][j];
                            }
                        }
                    }
                }
                //if(bestcurrentsaving > 0){
                    tour[nodeA][2] = nodeB+1;
                    tourcost[nodeA] += demand[nodeB];
                    intour = true;
                    active[nodeA] = true;
                    endA = nodeA;
                    endB = nodeB;
                    belongs[nodeB] = nodeA;
                /*}
                else{
                    terminate = true;
                }*/
            }
            else{
                boolean newnodefound = false;
                for(int i = 0 ; i < sm.length ; i++){
                    if(active[belongs[i]] && (i == endA || i == endB)){
                        for(int j = 0 ; j < sm.length ; j++){
                            if(!finished[belongs[j]] && !active[belongs[j]] && bestcurrentsaving < sm[i][j] && tourcost[belongs[i]] + demand[j] <= cap){
                                nodeA = i;
                                nodeB = j;
                                bestcurrentsaving = sm[i][j];
                                newnodefound = true;
                            }
                        }
                    }
                }
                //if(bestcurrentsaving > 0){
                if(newnodefound){
                    int[] newtour = new int[m.length+1];
                    int pos = 0;
                    if(nodeA==endB){
                        for(int j = 0 ; j <= m.length ; j++){
                            if(tour[belongs[nodeA]][j] == nodeA + 1){
                                pos = j;
                            }
                        }
                        tour[belongs[nodeA]][pos+1] = nodeB + 1;
                        tourcost[belongs[nodeA]] += demand[nodeB];
                        endB = nodeB;
                    }
                    else{
                        newtour[0] = 0;
                        newtour[1] = nodeB + 1;
                        for(int j = 2 ; j <= m.length ; j++){
                            newtour[j] = tour[belongs[nodeA]][j-1];
                        }
                        tour[belongs[nodeA]] = newtour;
                        tourcost[belongs[nodeA]] += demand[nodeB];
                        endA = nodeB;
                    }
                    belongs[nodeB] = belongs[nodeA];
                }
                else{
                    intour = false;
                    finished[belongs[nodeA]] = true;
                    active[belongs[nodeA]] = false;
                    carsused++;
                    if(carsused == travellers){
                        terminate = true;
                    }
                }
            }
        }
        int[] alltours = new int[m.length + travellers];
        alltours[m.length] = 0;
        int atpos = 0;
        for(int i = 0 ; i < sm.length ; i++){
            if(finished[i]){
                alltours[atpos] = 0;
                atpos++;
                int j = 1;
                while(tour[i][j] != 0){
                    alltours[atpos] = tour[i][j];
                    atpos++;
                    j++;
                }
            }
        }
        for(int j = 0 ; j < m.length + travellers; j++){
            if(j == 0){
                System.out.print("["+alltours[j]+" ");
            }
            else if(j == m.length + travellers - 1){
                System.out.println(alltours[j]+"]");
            }
            else{
                System.out.print(alltours[j]+" ");
            }
        }
        return alltours;
    }
    /** Runs Farthest Insertion algorithm based on input parameters */
    public int[] farthestInsertion(int[][] prem,int travellers){
        int[][] m = createExtendedCostMatrix(prem,travellers);
        int size = m.length + 1;
        int temp = 0;
        int pos = 0;
        int tourlength = 0;
        int included = 0;
        int minimax,loctemp,locpos,min;
        int[] tour = new int[size];
        boolean[] used = new boolean[m.length];
        used[0] = true;
        for(int i = 1 ; i < used.length ; i++){
            used[i] = false;
        }
        for(int i = 1 ; i < m.length ; i++){
            minimax = 0;
            for(int j = 1 ; j < m.length ; j++){
                if(!used[j]){
                    min = 2000000;
                    loctemp = 0;
                    locpos = 0;
                    for(int k = 0 ; k < i ; k++){
                        if(m[tour[k]][j] <= min){
                            min = m[tour[k]][j];
                            loctemp = j;
                            locpos = k;
                        }
                    }
                    if(min > minimax){
                        pos = locpos;
                        temp = loctemp;
                        minimax = min;
                    }
                }
            }
            if(tour[pos] < travellers){
                int bestlength = 2000000;
                for(int s = 0; s <= included ; s++){
                    int currentlength = tourlength;
                    if(tour[s] < travellers){
                        if(s > 0 && m[tour[s+1]][temp] + m[tour[s]][temp] - m[tour[s]][tour[s+1]] <= m[tour[s-1]][temp] + m[tour[s]][temp] - m[tour[s]][tour[s-1]]){
                            currentlength += m[tour[s+1]][temp] + m[tour[s]][temp] - m[tour[s]][tour[s+1]];
                        }
                        else if(s == 0){
                            if(m[tour[s+1]][temp] + m[tour[s]][temp] - m[tour[s]][tour[s+1]] <= m[tour[included]][temp] + m[tour[s]][temp] - m[tour[s]][tour[included]]){
                                currentlength += m[tour[s+1]][temp] + m[tour[s]][temp] - m[tour[s]][tour[s+1]];
                            }
                            else{
                                currentlength += m[tour[included]][temp] + m[tour[s]][temp] - m[tour[s]][tour[included]];
                            }
                        }
                        else{
                            currentlength += m[tour[s-1]][temp] + m[tour[s]][temp] - m[tour[s]][tour[s-1]];
                        }
                        if(currentlength < bestlength){
                            pos = s;
                            bestlength = currentlength;
                        }
                    }
                }
            }
            int[] newtour = new int[tour.length];
            if(pos > 0 && m[tour[pos+1]][temp] + m[tour[pos]][temp] - m[tour[pos]][tour[pos+1]] <= m[tour[pos-1]][temp] + m[tour[pos]][temp] - m[tour[pos]][tour[pos-1]]){
                for(int l = 0 ; l <= pos ; l++){
                    newtour[l] = tour[l];
                }
                newtour[pos+1] = temp;
                for(int l = pos + 2 ; l < tour.length ; l++){
                    newtour[l] = tour[l-1];
                }
                tourlength += m[tour[pos+1]][temp] + m[tour[pos]][temp] - m[tour[pos]][tour[pos+1]];
            }
            else if(pos == 0){
                if(m[tour[pos+1]][temp] + m[tour[pos]][temp] - m[tour[pos]][tour[pos+1]] <= m[tour[included]][temp] + m[tour[pos]][temp] - m[tour[pos]][tour[included]]){
                    newtour[pos+1] = temp;
                    for(int l = pos + 2 ; l < tour.length ; l++){
                        newtour[l] = tour[l-1];
                    }
                    tourlength += m[tour[pos+1]][temp] + m[tour[pos]][temp] - m[tour[pos]][tour[pos+1]];
                }
                else{
                    for(int l = 0 ; l <= included ; l++){
                        newtour[l] = tour[l];
                    }
                    newtour[included+1] = temp;
                    tourlength += m[tour[included]][temp] + m[tour[pos]][temp] - m[tour[pos]][tour[included]];
                }
            }
            else{
                for(int l = 0 ; l < pos ; l++){
                    newtour[l] = tour[l];
                }
                newtour[pos] = temp;
                for(int l = pos + 1 ; l < tour.length ; l++){
                    newtour[l] = tour[l-1];
                }
                tourlength += m[tour[pos-1]][temp] + m[tour[pos]][temp] - m[tour[pos]][tour[pos-1]];
            }
            used[temp] = true;
            included++;
            tour = newtour;
        }
        for(int i = 0 ; i < size ; i++){
            tour[i] = tour[i] - travellers + 1;
            if(tour[i] < 0){
                tour[i] = 0;
            }
        }
        for(int s = 0; s < tour.length ; s++){
            if(s == 0){
                System.out.print("[" + tour[s] + " ");
            }
            else if(s == tour.length - 1){
                System.out.println(tour[s] + "]");
            }
            else{
                System.out.print(tour[s] + " ");
            }
        }
        return tour;
    }
    /** Creates the extended cost matrix needed for Farthest Insertion */
    public int[][] createExtendedCostMatrix(int[][] costs, int travelers){
        int olength = costs.length;
        int nlength = olength + travelers - 1;
        int[] depotCosts = new int[olength-1];
        for(int i = 1 ; i <= depotCosts.length ; i++){
            depotCosts[i-1] = costs[0][i];
        }
       
        int[][] extendedmatrix = new int[nlength][nlength];
        for(int i = 0 ; i < travelers ; i++){
            extendedmatrix[i][i] = 0;
            for(int j = i + 1 ; j < travelers; j++){
                extendedmatrix[i][j] = 1000000;
                extendedmatrix[j][i] = 1000000;
            }
        }
        for(int i = travelers ; i < nlength ; i++){
            for(int j = 0 ; j < travelers ; j++){
                extendedmatrix[i][j] = depotCosts[i-travelers];
                extendedmatrix[j][i] = depotCosts[i-travelers];
            }
        }
        for(int i = travelers ; i < nlength ; i++){
            extendedmatrix[i][i] = 0;
            for(int j = i + 1 ; j < nlength ; j++){
                extendedmatrix[i][j] = costs[i-travelers+1][j-travelers+1];
                extendedmatrix[j][i] = costs[i-travelers+1][j-travelers+1];
            }
        }
        return extendedmatrix;
    }
    
    /** Generates test matrix for the fixed optimisation test */
    public int[][] generateTestMatrix(){
        int[][] out = new int[16][16];
        for(int i = 0 ; i < 16 ; i++){
            out[i][i] = 0;
            for(int j = i + 1 ; j < 16 ; j++){
                if((i==2 || i==3 || i==7 || i==8 || i==12 || i==13) && j==i+1){
                    out[i][j] = 3;
                    out[j][i] = 3;
                }
                else if(((i==1 || i==4 || i==6 || i==9 || i==11 || i==14) && j==i+1) || ((i==1 || i==6 || i==11) && j==i+4)){
                    out[i][j] = 4;
                    out[j][i] = 4;
                }
                else if(i==0 && (j==1 || j==5 || j==6 || j==10 || j==11 || j==15)){
                    out[i][j] = 5;
                    out[j][i] = 5;
                }
                else if((i==1 || i==2 || i==3 || i==6 || i==7 || i==8 || i==11 || i==12 || i==13) && j==i+2){
                    out[i][j] = 6;
                    out[j][i] = 6;
                }
                else if((i==1 || i==2 || i==6 || i==7 || i==11 || i==12) && j==i+3){
                    out[i][j] = 7;
                    out[j][i] = 7;
                }
                else if(i==0 && (j==2 || j==4 || j==7 || j==9 || j==12 || j==14)){
                    out[i][j] = 8;
                    out[j][i] = 8;
                }
                else if((i==1 && j==15) || ((i==5 || i==10) && j==i+1)){
                    out[i][j] = 9;
                    out[j][i] = 9;
                }
                else if(i==0 && (j==3 || j==8 || j==13)){
                    out[i][j] = 10;
                    out[j][i] = 10;
                }
                else if((i==2 && j==15) || ((i==4 || i==9) && j==i+2)){
                    out[i][j] = 11;
                    out[j][i] = 11;
                }
                else if(((i==1 || i==6) && (j==i+5 || j==i+13)) || ((i==5 || i==10) && (j==i+2 || j==i+10)) || (i==3 && j==15)){
                    out[i][j] = 12;
                    out[j][i] = 12;
                }
                else if(((i==3 || i==8 || i==13) && j==i+3) || ((i==5 || i==10) && j==i+5) || (i==1 && j==11)){
                    out[i][j] = 13;
                    out[j][i] = 13;
                }
                else if(((i==1 || i==6) && (j==i+6 || j==i+9 || j==i+12)) || ((i==2 || i==7) && (j==i+4 || j==i+12)) || ((i==4 || i==9) && (j==i+3 || j==i+11)) || ((i==5 || i==10) && (j==i+3 || j==i+6 || j==i+9))){
                    out[i][j] = 14;
                    out[j][i] = 14;
                }
                else if((i==1 && j==12) || (i==2 && j==11) || (i==3 && j==14) || ((i==4 || i==9) && j==i+6) || ((i==5 || i==10) && j==i+4)){
                    out[i][j] = 15;
                    out[j][i] = 15;
                }
                else if(((i==1 || i==6) && j==i+7) || (i==2 && j==13) || ((i==3 || i==8) && (j==i+4 || j==i+7 || j==i+8)) || ((i==4 || i==9) && j==i+4) || (i==5 && j==13)){
                    out[i][j] = 16;
                    out[j][i] = 16;
                }
                else if(i==1 || i==6 || ((i==2 || i==7) && (j==i+5 || j==i+8 || j==i+10)) || (i==3 && j==13) || ((i==4 || i==9) && j==i+5) || i==5){
                    out[i][j] = 17;
                    out[j][i] = 17;
                }
                else if((i==3 || i==8) && (j==i+5 || j==i+9)){
                    out[i][j] = 18;
                    out[j][i] = 18;
                }
                else if(i==3 || i==8 || ((i==2 || i==7) && (j==i+6)) || (i==4 && j==13)){
                    out[i][j] = 19;
                    out[j][i] = 19;
                }
                else{
                    out[i][j] = 20;
                    out[j][i] = 20;
                }
            }
        }
        return out;
    }
}
